
{*******************************************************}
{                                                       }
{                  SnmpEye Version 1.3                  }
{                                                       }
{                                                       }
{          Copyright (c) 1999-2004 Vadim Crits          }
{                                                       }
{*******************************************************}

unit SeMain;

interface

uses
  Windows, Messages, SysUtils, Classes, Graphics, Controls, Forms, Dialogs,
  ExtCtrls, ComCtrls, ToolWin, Menus, ActnList, StdCtrls, ImgList, Registry,
  Snmp;

type
  TMainForm = class(TForm)
    StatusBar: TStatusBar;
    ActionList: TActionList;
    actSave: TAction;
    actSaveAs: TAction;
    actExit: TAction;
    actToolBar: TAction;
    actStatusBar: TAction;
    actGet: TAction;
    actGetNext: TAction;
    actAbout: TAction;
    actWalk: TAction;
    ListView: TListView;
    ImageList: TImageList;
    actRun: TAction;
    actClear: TAction;
    actOptions: TAction;
    CoolBar: TCoolBar;
    psMenuBar: TPageScroller;
    MenuBar: TToolBar;
    btnFile: TToolButton;
    btnView: TToolButton;
    btnAct: TToolButton;
    btnHelp: TToolButton;
    actDetails: TAction;
    actStop: TAction;
    MainMenu: TMainMenu;
    miFile: TMenuItem;
    miView: TMenuItem;
    miAction: TMenuItem;
    miHelp: TMenuItem;
    miSave: TMenuItem;
    miSaveAs: TMenuItem;
    N1: TMenuItem;
    miExit: TMenuItem;
    miToolbar: TMenuItem;
    miStatusBar: TMenuItem;
    N2: TMenuItem;
    miOptions: TMenuItem;
    miDetails: TMenuItem;
    N3: TMenuItem;
    miRun: TMenuItem;
    miStop: TMenuItem;
    miClear: TMenuItem;
    N4: TMenuItem;
    miGet: TMenuItem;
    miGetNext: TMenuItem;
    miWalk: TMenuItem;
    miAbout: TMenuItem;
    Splitter: TSplitter;
    btnRun: TToolButton;
    btnStop: TToolButton;
    btnClear: TToolButton;
    lblAgent: TLabel;
    cboAgent: TComboBox;
    lblCommunity: TLabel;
    cboCommunity: TComboBox;
    lblOids: TLabel;
    cboOids: TComboBox;
    psToolBar: TPageScroller;
    ToolBar: TToolBar;
    procedure FormCreate(Sender: TObject);
    procedure FormDestroy(Sender: TObject);
    procedure ListViewDblClick(Sender: TObject);
    procedure ListViewKeyDown(Sender: TObject; var Key: Word;
      Shift: TShiftState);
    procedure StatusBarHint(Sender: TObject);
    procedure actSaveExecute(Sender: TObject);
    procedure actSaveAsExecute(Sender: TObject);
    procedure actExitExecute(Sender: TObject);
    procedure actToolBarExecute(Sender: TObject);
    procedure actStatusBarExecute(Sender: TObject);
    procedure actDetailsExecute(Sender: TObject);
    procedure actOptionsExecute(Sender: TObject);
    procedure actRunExecute(Sender: TObject);
    procedure actStopExecute(Sender: TObject);
    procedure actClearExecute(Sender: TObject);
    procedure ActionExecute(Sender: TObject);
    procedure actAboutExecute(Sender: TObject);
    procedure ActionListUpdate(Action: TBasicAction; var Handled: Boolean);
    procedure SplitterCanResize(Sender: TObject; var NewSize: Integer;
      var Accept: Boolean);
    procedure FormClose(Sender: TObject; var Action: TCloseAction);
  private
    { Private declarations }
    FDone: Boolean;
    FAction: Integer;
    FFileName: string;
    procedure ReadHistoryList(Reg: TRegistry; List: TStrings);
    procedure WriteHistoryList(Reg: TRegistry; List: TStrings);
    procedure UpdateList(List: TComboBox);
    procedure CreateVariableBindings(const Oids: string;
      var VarBindList: TSnmpVarBindList);
    procedure PrintVarBind(const VarBind: TSnmpVarBind);
    procedure MkFile(const FileName: string);
  public
    { Public declarations }
  end;

var
  MainForm: TMainForm;
  nTimeOut: Integer = 1000;
  nRetries: Integer = 1;

implementation

uses SeOptions, SeDetails, SeAbout, SeConst, Mgmtapi;

{$R *.dfm}

procedure TMainForm.CreateVariableBindings(const Oids: string;
  var VarBindList: TSnmpVarBindList);
var
  Position: DWORD;
  i: Integer;
  VarBind: TSnmpVarBind;
begin
  with TStringList.Create do
  try
    Text := StringReplace(Oids, SOidsSep, sLineBreak, [rfReplaceAll]);
    VarBindList.list := SnmpUtilMemAlloc(SizeOf(TSnmpVarBind) * Count);
    VarBindList.len := Count;
    Position := 0;
    try
      for i := 0 to Count - 1 do
      begin
        if not SnmpMgrStrToOid(PChar(Strings[i]), @VarBind.name) then
          raise Exception.CreateFmt(SInvalidOid, [Strings[i]]);
        VarBind.value.asnType := ASN_NULL;
        System.Move(VarBind, Pointer(DWORD(VarBindList.list) + Position)^, SizeOf(TSnmpVarBind));
        Inc(Position, SizeOf(TSnmpVarBind));
      end;
     except
       SnmpUtilVarBindListFree(@VarBindList);
       raise;
     end;
  finally
    Free;
  end;
end;

procedure TMainForm.PrintVarBind(const VarBind: TSnmpVarBind);
var
  A: array[0..3] of string;
  Str: PChar;
  Secs, Days, IpAddr: TAsnUnsigned32;
  ts: TTimeStamp;
  dt: TDateTime;
  IsAscii: Boolean;
  P: PByte;
  i: Integer;
begin
  A[0] := '.' + SnmpUtilOidToA(@VarBind.name);
  SnmpMgrOidToStr(@VarBind.name, @Str);
  try
    SetLength(A[1], StrLen(Str));
    StrCopy(PChar(A[1]), Str);
  finally
    SnmpUtilMemFree(Str);
  end;
  with VarBind.value do
    case asnType of
      ASN_NULL:
        A[2] := 'Null';
      ASN_INTEGER:
        begin
          A[2] := 'Integer32';
          A[3] := Format('%d', [number]);
        end;
      ASN_COUNTER32:
        begin
          A[2] := 'Counter32';
          A[3] := Format('%u', [counter]);
        end;
      ASN_GAUGE32:
        begin
          A[2] := 'Gauge32';
          A[3] := Format('%u', [gauge]);
        end;
      ASN_UNSIGNED32:
        begin
          A[2] := 'Unsigned32';
          A[3] := Format('%u', [unsigned32]);
        end;
      ASN_COUNTER64:
        begin
          A[2] := 'Counter64';
          A[3] := Format('%u', [counter64.QuadPart]);
        end;
      ASN_TIMETICKS:
        begin
          Secs := ticks div 100;
          Days := 0;
          if Secs >= SecsPerDay then
          begin
            Days := Secs div SecsPerDay;
            Secs := Secs mod SecsPerDay;
          end;
          ts.Time := Secs * 1000;
          ts.Date := DateDelta;
          dt := TimeStampToDateTime(ts);
          A[2] := 'TimeTicks';
          A[3] := Format('%u (%u days, %s hours, %s minutes, %s seconds)',
                         [ticks, Days, FormatDateTime('h', dt),
                         FormatDateTime('n', dt), FormatDateTime('s', dt)]);
        end;
      ASN_IPADDRESS:
        begin
          IpAddr := PDWORD(address.stream)^;
          A[2] := 'IpAddress';
          A[3] := Format('%u.%u.%u.%u', [IpAddr and $FF,
                                        (IpAddr shr 8) and $FF,
                                        (IpAddr shr 16) and $FF,
                                        (IpAddr shr 24) and $FF]);
        end;
      ASN_OCTETSTRING:
        begin
          A[2] := 'Octet String';
          if str.length > 0 then
          begin
            IsAscii := True;
            P := str.stream;
            for i := 0 to str.length - 1 do
            begin
              if not (P^ in [0, 32..127]) then
              begin
                IsAscii := False;
                Break;
              end;
              Inc(P);
            end;
            if IsAscii then
              A[3] := Copy(PChar(str.stream), 1, str.length)
            else
            begin
              P := str.stream;
              for i := 0 to str.length - 1 do
              begin
                A[3] := A[3] + Format('%2.2x ', [P^]);
                Inc(P);
              end;
            end;
          end;
        end;
      ASN_BITS, ASN_SEQUENCE, ASN_OPAQUE:
        begin
          if asnType = ASN_BITS then
            A[2] := 'Bits'
          else if asnType = ASN_SEQUENCE then
            A[2] := 'Sequence'
          else
            A[2] := 'Opaque';
          if bits.length > 0 then
          begin
            P := bits.stream;
            for i := 0 to bits.length - 1 do
            begin
              A[3] := A[3] + Format('%2.2x ', [P^]);
              Inc(P);
            end;
          end;
        end;
      ASN_OBJECTIDENTIFIER:
        begin
          A[2] := 'Object Identifier';
          A[3] := '.' + SnmpUtilOidToA(@obj);
        end;
      else
        A[2] := 'Unexpected Type';
    end;
  with ListView.Items.Add do
  begin
    ImageIndex := 0;
    Caption := A[0];
    SubItems.Add(A[1]);
    SubItems.Add(A[2]);
    SubItems.Add(A[3]);
  end;
end;

procedure TMainForm.FormCreate(Sender: TObject);
var
  Reg: TRegistry;
  Position: TRect;
begin
  FAction := 1;
  Reg := TRegistry.Create;
  with Reg do
  try
    if OpenKey(INIT_KEY, False) then
    try
      if ValueExists(SPosition) then
      begin
        ReadBinaryData(SPosition, Position, SizeOf(Position));
        BoundsRect := Position;
      end;
      if ValueExists(SWindowState) then
        if ReadInteger(SWindowState) = Ord(wsMaximized) then
          WindowState := wsMaximized;
      if ValueExists(SShowToolBar) then
        psToolBar.Visible := ReadBool(SShowToolBar);
      if ValueExists(SShowStatusBar) then
        StatusBar.Visible := ReadBool(SShowStatusBar);
      if ValueExists(SAction) then
        FAction := ReadInteger(SAction);
      if ValueExists(STimeOut) then
        nTimeOut := ReadInteger(STimeOut);
      if ValueExists(SRetries) then
        nRetries := ReadInteger(SRetries);
    finally
      CloseKey;
    end;
    if OpenKey(HL_AGENT_KEY, False) then
    try
      ReadHistoryList(Reg, cboAgent.Items);
      cboAgent.ItemIndex := 0;
     finally
      CloseKey;
    end;
    if OpenKey(HL_COMMUNITY_KEY, False) then
    try
      ReadHistoryList(Reg, cboCommunity.Items);
      cboCommunity.ItemIndex := 0;
    finally
      CloseKey;
    end;
    if OpenKey(HL_OID_KEY, False) then
    try
      ReadHistoryList(Reg, cboOids.Items);
      cboOids.ItemIndex := 0;
    finally
      CloseKey;
    end
    else
    begin
      cboAgent.ItemIndex := 0;
      cboCommunity.ItemIndex := 0;
      cboOids.ItemIndex := 0;
    end;
  finally
    Reg.Free;
  end;
  FDone := True;
  StatusBar.SimpleText := SReady;
end;

procedure TMainForm.FormDestroy(Sender: TObject);
var
  Reg: TRegistry;
  Position: TRect;
begin
  Reg := TRegistry.Create;
  with Reg do
  try
    if OpenKey(INIT_KEY, True) then
    try
      if WindowState = wsNormal then
      begin
        Position := BoundsRect;
        WriteBinaryData(SPosition, Position, SizeOf(Position));
      end;
      WriteInteger(SWindowState, Ord(WindowState));
      WriteBool(SShowToolBar, psToolBar.Visible);
      WriteBool(SShowStatusBar, StatusBar.Visible);
      WriteInteger(SAction, FAction);
      WriteInteger(STimeOut, nTimeOut);
      WriteInteger(SRetries, nRetries);
    finally
      CloseKey;
    end;
    if OpenKey(HL_AGENT_KEY, True) then
    try
      WriteHistoryList(Reg, cboAgent.Items);
    finally
      CloseKey;
    end;
    if OpenKey(HL_COMMUNITY_KEY, True) then
    try
      WriteHistoryList(Reg, cboCommunity.Items);
    finally
      CloseKey;
    end;
    if OpenKey(HL_OID_KEY, True) then
    try
      WriteHistoryList(Reg, cboOids.Items);
    finally
      CloseKey;
    end;
  finally
    Reg.Free;
  end;
end;

procedure TMainForm.ListViewDblClick(Sender: TObject);
begin
  actDetails.Execute;
end;

procedure TMainForm.ListViewKeyDown(Sender: TObject; var Key: Word;
  Shift: TShiftState);
begin
  if Key = VK_RETURN then actDetails.Execute;
end;

procedure TMainForm.StatusBarHint(Sender: TObject);
var
  SHint: string;
begin
  SHint := Application.Hint;
  with StatusBar do
    if SHint <> '' then
    begin
      SimplePanel := False;
      Panels[0].Text := SHint;
    end
    else
    begin
      SimplePanel := True;
      SimpleText := SReady;
    end;
end;

procedure TMainForm.actSaveExecute(Sender: TObject);
begin
  if FFileName = '' then
    actSaveAs.Execute
  else
    MkFile(FFileName);
end;

procedure TMainForm.actSaveAsExecute(Sender: TObject);
var
  FileExt: string;
begin
  Application.ProcessMessages;
  with TSaveDialog.Create(Self) do
  try
    Options := [ofHideReadOnly, ofEnableSizing, ofOverwritePrompt];
    if FFileName = '' then FileName := '*' + SDefExt else FileName := FFileName;
    Filter := SFilter;
    if Execute then
    begin
      FFileName := FileName;
      FileExt := ExtractFileExt(FFileName);
      if AnsiLowerCase(FileExt) <> SDefExt then
      begin
        Delete(FFileName, Pos('.', FFileName), Length(FileExt));
        FFileName := FFileName + SDefExt;
      end;
      MkFile(FFileName);
    end;
  finally
    Free;
  end;
end;

procedure TMainForm.actExitExecute(Sender: TObject);
begin
  Close;
end;

procedure TMainForm.actToolBarExecute(Sender: TObject);
begin
  psToolBar.Visible := not psToolBar.Visible;
end;

procedure TMainForm.actStatusBarExecute(Sender: TObject);
begin
  StatusBar.Visible := not StatusBar.Visible;
end;

procedure TMainForm.actDetailsExecute(Sender: TObject);
begin
  ShowDetailsDialog(Self);
end;

procedure TMainForm.actOptionsExecute(Sender: TObject);
begin
  with TOptionsDialog.Create(Self) do
  try
    udTimeOut.Position := nTimeOut;
    udRetries.Position := nRetries;
    ShowModal;
    if ModalResult = mrOk then
    begin
      nTimeout := udTimeOut.Position;
      nRetries := udRetries.Position;
    end;
  finally
    Free;
  end;
end;

procedure TMainForm.actRunExecute(Sender: TObject);
var
  SaveCursor: TCursor;
  Session: PSnmpMgrSession;
  VarBindList: TSnmpVarBindList;
  ErrorStatus, ErrorIndex: TAsnInteger32;
  MultipleOids, AllOids: Boolean;
  Oid, RootOid: TAsnObjectIdentifier;
  RequestType: Byte;
  RetVal, i: Integer;
  Position: DWORD;
begin
  SaveCursor := Screen.Cursor;
  FDone := False;
  (Sender as TAction).Update;
  with StatusBar do
  begin
    AutoHint := False;
    SimplePanel := True;
    SimpleText := SUpdating;
  end;
  Application.ProcessMessages;
  try
    Screen.Cursor := crAppStart;
    if (Trim(cboAgent.Text) = '') or (Trim(cboCommunity.Text) = '') or
      (Trim(cboOids.Text) = '') then raise Exception.Create(SEmptyParams);
    Session := SnmpMgrOpen(PChar(cboAgent.Text), PChar(cboCommunity.Text), nTimeOut, nRetries);
    if not Assigned(Session) then
      raise Exception.CreateFmt(SSnmpMgrOpenErr, [GetLastError]);
    try
      MultipleOids := Pos(SOidsSep, cboOids.Text) > 0;
      AllOids := AnsiSameText(cboOids.Text, SAllOids);
      if (FAction = 1) and MultipleOids then
        CreateVariableBindings(cboOids.Text, VarBindList)
      else if (FAction = 3) and AllOids then
        SnmpMgrStrToOid('1', @Oid)
      else if not SnmpMgrStrToOid(PChar(cboOids.Text), @Oid) then
        raise Exception.CreateFmt(SInvalidOid, [cboOids.Text]);
      UpdateList(cboAgent);
      UpdateList(cboCommunity);
      UpdateList(cboOids);
      if not MultipleOids then
        with VarBindList do
        begin
          list := SnmpUtilMemAlloc(SizeOf(TSnmpVarBind));
          len := 1;
          with list^ do
          begin
            name := Oid;
            value.asnType := ASN_NULL;
          end;
        end;
      try
        case FAction of
          1, 2:
            begin
              if FAction = 1 then
                RequestType := SNMP_PDU_GET
              else
                RequestType := SNMP_PDU_GETNEXT;
              RetVal := SnmpMgrRequest(Session, RequestType,
                                       @VarBindList, @ErrorStatus, @ErrorIndex);
              if RetVal = SNMPAPI_ERROR then
                raise Exception.CreateFmt(SSnmpMgrRequestErr, [GetLastError])
              else if ErrorStatus <> SNMP_ERRORSTATUS_NOERROR then
                raise Exception.CreateFmt(SSnmpMgrRequestErrStatus,
                                          [ErrorStatus, ErrorIndex])
              else if MultipleOids then
              begin
                Position := 0;
                for i := 0 to VarBindList.len - 1 do
                begin
                  PrintVarBind(PSnmpVarBind(DWORD(VarBindList.list) + Position)^);
                  Inc(Position, SizeOf(TSnmpVarBind));
                end;
              end
              else
                PrintVarBind(VarBindList.list^);
            end;
          3:
            begin
              RequestType := SNMP_PDU_GETNEXT;
              if not AllOids then
                SnmpUtilOidCpy(@RootOid, @VarBindList.list^.name);
              while True do
              begin
                RetVal := SnmpMgrRequest(Session, RequestType,
                                         @VarBindList, @ErrorStatus, @ErrorIndex);
                if RetVal = SNMPAPI_ERROR then
                  raise Exception.CreateFmt(SSnmpMgrRequestErr, [GetLastError]);
                if not AllOids then
                begin
                  if (ErrorStatus = SNMP_ERRORSTATUS_NOSUCHNAME) or
                    (SnmpUtilOidNCmp(@RootOid, @VarBindList.list^.name, RootOid.idLength) <> 0) then
                    Break;
                end
                else if ErrorStatus = SNMP_ERRORSTATUS_NOSUCHNAME then Break;
                if ErrorStatus <> SNMP_ERRORSTATUS_NOERROR then
                  raise Exception.CreateFmt(SSnmpMgrRequestErrStatus,
                                            [ErrorStatus, ErrorIndex]);
                PrintVarBind(VarBindList.list^);
                if FDone then Break;
                Application.ProcessMessages;
              end;
            end;
        end;
      finally
        SnmpUtilVarBindListFree(@VarBindList);
      end;
    finally
      SnmpMgrClose(Session);
    end;
  finally
    FDone := True;
    with StatusBar do
    begin
      SimpleText := SReady;
      AutoHint := True;
    end;
    Screen.Cursor := SaveCursor;
  end;
end;

procedure TMainForm.actStopExecute(Sender: TObject);
begin
  FDone := True;
  Application.ProcessMessages;
end;

procedure TMainForm.actClearExecute(Sender: TObject);
begin
  with ListView.Items do
    if Count > 0 then
    begin
      BeginUpdate;
      Clear;
      EndUpdate;
    end;
end;

procedure TMainForm.ActionExecute(Sender: TObject);
begin
  FAction := (Sender as TAction).Tag;
end;

procedure TMainForm.actAboutExecute(Sender: TObject);
begin
  ShowAboutBox(Self);
end;

procedure TMainForm.ActionListUpdate(Action: TBasicAction;
  var Handled: Boolean);
begin
  actRun.Enabled := FDone;
  actSave.Enabled := actRun.Enabled and (ListView.Items.Count > 0);
  actSaveAs.Enabled := actSave.Enabled;
  actToolBar.Checked := psToolBar.Visible;
  actStatusBar.Checked := StatusBar.Visible;
  actDetails.Enabled := Assigned(ListView.Selected);
  actStop.Enabled := not actRun.Enabled;
  actClear.Enabled := actSave.Enabled;
  if (Action as TAction).GroupIndex = 1 then
    (Action as TAction).Checked := (Action as TAction).Tag = FAction;
end;

procedure TMainForm.UpdateList(List: TComboBox);
var
  i: Integer;
begin
  with List do
  begin
    if Items.Count > 0 then
      for i := 0 to Items.Count - 1 do
        if AnsiSameText(Items[i], Text) then
        begin
          Items.Move(i, 0);
          ItemIndex := 0;
          Exit;
        end;
    Items.Insert(0, Text);
    if Items.Count > 10 then
      Items.Delete(Items.Count - 1);
  end;
end;

procedure TMainForm.ReadHistoryList(Reg: TRegistry; List: TStrings);
var
  sl: TStrings;
  i: Integer;
begin
  sl := TStringList.Create;
  try
    Reg.GetValueNames(sl);
    List.Clear;
    for i := 0 to sl.Count - 1 do
      List.Add(Reg.ReadString(sl[i]));
  finally
    sl.Free;
  end;
end;

procedure TMainForm.WriteHistoryList(Reg: TRegistry; List: TStrings);
var
  i: Integer;
begin
  for i := 0 to List.Count - 1 do
    Reg.WriteString('Item' + IntToStr(i), List[i]);
end;

procedure TMainForm.MkFile(const FileName: string);
var
  F: TextFile;
  SaveCursor: TCursor;
  i: Integer;
begin
  AssignFile(F, FileName);
  Rewrite(F);
  SaveCursor := Screen.Cursor;
  try
    Screen.Cursor := crAppStart;
    with ListView do
      for i := 0 to Items.Count - 1 do
        Writeln(F, Format('%s | %s | %s | %s', [Items[i].Caption,
                                                Items[i].SubItems[0],
                                                Items[i].SubItems[1],
                                                Items[i].SubItems[2]]));
  finally
    CloseFile(F);
    Screen.Cursor := SaveCursor;
  end;
end;

procedure TMainForm.SplitterCanResize(Sender: TObject;
  var NewSize: Integer; var Accept: Boolean);
var
  i: Integer;
begin
  with CoolBar do
    if NewSize < Height then
    begin
      for i := Bands.Count - 1 downto 0 do
        if Bands[i].Break = True then
          if NewSize <= Height - Bands[i].Height then
          begin
            Bands[i].Break := False;
            Accept := True;
          end;
    end
    else
    begin
      for i := Bands.Count - 1 downto 0 do
        if (Bands[i].Break = False) and (not Bands[i].FixedSize) then
          if NewSize >= Height + Bands[i].Height then
          begin
            Bands[i].Break := True;
            Accept := True;
          end;
    end;
end;

procedure TMainForm.FormClose(Sender: TObject; var Action: TCloseAction);
begin
  if actStop.Enabled then actStop.Execute;
end;

end.
